@@ -40,17 +40,17 @@ sub new
     return $self;
 }
 
-sub setVerbosity     { my ($self, $verbosity) = @_; $self->{verbosity} = defined($verbosity) ? $verbosity : 0;   }
-sub symboltable      { my ($self) = @_; return $self->{symboltable};                                             }
-sub setoutputfile    { my ($self, $outputfilename) = @_; $self->{outputfilename} = $outputfilename;              }
+sub setVerbosity         { my ($self, $verbosity) = @_; $self->{verbosity} = defined($verbosity) ? $verbosity : 0;   }
+sub symboltable          { my ($self) = @_; return $self->{symboltable};                                             }
+sub setoutputfile        { my ($self, $outputfilename) = @_; $self->{outputfilename} = $outputfilename;              }
 
-sub fragment2class   { my ($self) = @_; $self->{generationoptions}{fragment2class} = 'true';                     }
-sub embedactions     { my ($self) = @_; $self->{generationoptions}{embedactions} = 'true';                       }
-sub stripactions     { my ($self) = @_; $self->{generationoptions}{stripactions} = 'true';                       }
-sub stripallcomments { my ($self) = @_; $self->{generationoptions}{stripallcomments} = 'true';                   }
-sub splitliteral     { my ($self) = @_; $self->{generationoptions}{splitliteral} = 'true';                       }
-sub shiftlazytogreedy{ my ($self) = @_; $self->{generationoptions}{shiftlazytogreedy} = 'true';                  }
-sub buildkeywords    { my ($self) = @_; $self->{generationoptions}{buildkeywords} = 'true';                      }
+sub fragment2class       { my ($self) = @_; $self->{generationoptions}{fragment2class} = 'true';                     }
+sub embedactions         { my ($self) = @_; $self->{generationoptions}{embedactions} = 'true';                       }
+sub stripactions         { my ($self) = @_; $self->{generationoptions}{stripactions} = 'true';                       }
+sub stripallcomments     { my ($self) = @_; $self->{generationoptions}{stripallcomments} = 'true';                   }
+sub matchcaseinsensitive { my ($self) = @_; $self->{generationoptions}{matchcaseinsensitive} = 'true';               }
+sub shiftlazytogreedy    { my ($self) = @_; $self->{generationoptions}{shiftlazytogreedy} = 'true';                  }
+sub buildkeywords        { my ($self) = @_; $self->{generationoptions}{buildkeywords} = 'true';                      }
 
 sub testoption
 {
@@ -301,7 +301,7 @@ sub translateLiteralCase
 }
 
 ##
-#   generateSplitLiteral : make a literal (somewhat) case-insensitive by creating a subrule with 2 match alternatives :
+#   generatematchcaseinsensitive : make a literal (somewhat) case-insensitive by creating a subrule with 2 match alternatives :
 #
 #   - all uppercase
 #   - characterclass with the 1st character in upper/lower case, the rest all lowercase
@@ -313,7 +313,7 @@ sub translateLiteralCase
 ##  CAVEAT : introduction of the ':i' and ':ic' literal and class modifiers has made this function obsolete. ##
 ##           we retain it as a demo of generated subrules.                                                   ##
 ## --------------------------------------------------------------------------------------------------------- ##
-sub generateSplitLiteral
+sub generateMatchCaseInsensitive
 {
     my ($tokenvalue) = @_;
 
@@ -405,7 +405,7 @@ sub convertRangeToClass
     $begr = $begr->{value};
     $endr = $endr->{value};
 
-    return { type => 'unsupported', msg => "can't convert range ${begr} .. ${endr} to class" } if !isSingleChar($begr) || !isSingleChar($endr);
+    return { type => 'unsupported', msg => "can't convert range (${begr} .. ${endr}) to class" } if !isSingleChar($begr) || !isSingleChar($endr);
 
     my $classtext = "${begr}-${endr}";
 
@@ -557,7 +557,7 @@ sub walktoken
         };
         (ref $token eq "HASH" && exists $token->{type} && $token->{type} eq "literal") && do {
             my $tokenvalue = $token->{value};
-            # if ( !$self->testoption('splitliteral') || !$self->isAlphaLiteral($tokenvalue) )
+            # if ( !$self->testoption('matchcaseinsensitive') || !$self->isAlphaLiteral($tokenvalue) )
             # {
                 $mxToken = { type => 'literal', value => $tokenvalue, isLexeme => 1 };
                 my $class4list = convertLiteralToClass($tokenvalue);
@@ -565,11 +565,11 @@ sub walktoken
             # }
             # else
             # {
-            #     my $literaltoken = generateSplitLiteral($tokenvalue);
-            #     delete $self->{generationoptions}{splitliteral};
+            #     my $literaltoken = generatematchcaseinsensitive($tokenvalue);
+            #     delete $self->{generationoptions}{matchcaseinsensitive};
             #     $mxToken = $self->walkgroup($rulename, $literaltoken, $context);
             #     $mxToken = $self->createSubRule( $rulename, $mxToken, $negation, $cardinality, $context );
-            #     $self->{generationoptions}{splitliteral} = 'true';
+            #     $self->{generationoptions}{matchcaseinsensitive} = 'true';
             # }
             last SWITCH;
         };
@@ -663,7 +663,7 @@ sub walknonterminal
 ## -------------------------------------------------------------------------------
 # 'retagSimpleRule' : check if the cardinality can be added to 'mxtoken'.
 #  Rules  :
-#  - '?' is consumed by 'opt_' subrules, so we reject it here
+#  - '?' is consumed by 'opt_' subrules, so we have to reject it here
 #  - 'mxtoken' must be a subrule of 'rule'
 #  - the parent rule must consist of a single element
 #    (responsibility of the caller to verify)
@@ -908,12 +908,13 @@ sub processRightSides
                 ##
                 #   replace a keyword built from lower/upper case character classes by a literal
                 ##
-                if ($self->testoption("buildkeywords") && $keywordfragment->{status} == 0 && $keywordfragment->{value} =~ /^[a-z]/i)
+                my $keyword = $keywordfragment->{value};
+                if ($self->testoption("buildkeywords") && $keywordfragment->{status} == 0 && $keyword =~ /^[a-z]/i)
                 {
                     $currentlist = [
-                        { type => "literal", value => $keywordfragment->{value}, isLexeme => 1 }
+                        { type => "literal", value => $keyword, isLexeme => 1 }
                     ];
-                    $newlist = { list => $currentlist };
+                    $newlist = { list => $currentlist, isLexeme => 1 };
                 }
                 else
                 {
@@ -1058,6 +1059,7 @@ sub processRightSides
         $currentlist = [
             { type => "literal", value => $keywordfragment->{value}, isLexeme => 1 }
         ];
+        $groupisLexeme->{status} = 0;
     }
 
     ##
@@ -1075,7 +1077,7 @@ sub processRightSides
     ##
     #   remove branches that are case-equivalent
     ##
-    $alternativelists = $self->removeCaseEquivalentBranches($alternativelists) if $self->testoption('splitliteral');
+    $alternativelists = $self->removeCaseEquivalentBranches($alternativelists) if $self->testoption('matchcaseinsensitive');
 
     my $grouplist = { type => 'group', list => $alternativelists };
     # single lists of literals can be tagged as lexemes
@@ -1108,7 +1110,7 @@ sub walkrule
 {
     my ($self, $rulename, $rule, $context) = @_;
 
-    # if ( $rulename eq "PROGRAM")
+    # if ( $rulename eq "ACCEPT")
     # {
     #     printf "found!\n";
     # }
@@ -1371,7 +1373,7 @@ sub processRedirected
 
     my $rule = $self->getMarpaRule(undef, $token);
 
-    if ($self->testoption('splitliteral') && $self->isKeywordLetter($token, $rule))
+    if ($self->testoption('matchcaseinsensitive') && $self->isKeywordLetter($token, $rule))
     {
         my $value = $rule->{class4list};
         $value = substr($value, 0, 1);
@@ -1397,7 +1399,7 @@ sub fragmentEligible2Convert
     return length($class4list) > 1;
 }
 
-sub convertFragment2Class
+sub writeFragmentAsClass
 {
     my ($self, $rulename, $rule, $options ) = @_;
 
@@ -1540,7 +1542,7 @@ sub computeRhs
         (exists $rhs->{type} && $rhs->{type} eq "class") && do {
             my $value = $rhs->{value};
             my $modifier = "";
-            if ($self->testoption('splitliteral') && $self->isAlphaLiteral($value))
+            if ($self->testoption('matchcaseinsensitive') && $self->isAlphaLiteral($value))
             {
                 my $normclass = $self->normalizeClassString($value);
                 $value = $normclass->{classstring};
@@ -1552,7 +1554,7 @@ sub computeRhs
         (exists $rhs->{type} && $rhs->{type} eq "literal") && do {
             my $literal = $rhs->{value};
             $literal =  ($literal eq "\\'") ? "[']" : "'${literal}'";
-            if ($self->testoption('splitliteral') && $self->isAlphaLiteral($literal))
+            if ($self->testoption('matchcaseinsensitive') && $self->isAlphaLiteral($literal))
             {
                 $literal = lc $literal;
                 $literal .= ':i';
@@ -1602,8 +1604,8 @@ sub printRhs
 
     mydie "printRhs: 'status' not defined for rule $rulename" if !exists $options->{status};
 
-    my $status  = $options->{status};
-    my $rhs     = $self->computeRhs($rulename, $rule, $options);
+    my $status          = $options->{status};
+    my $rhs             = $self->computeRhs($rulename, $rule, $options);
 
     return "" if exists $rhs->{discard};
 
@@ -1798,7 +1800,7 @@ sub writeMarpaRule
 
     $self->tagMarpaRule($rulename);
 
-    # if ($rulename =~ /HEXNUMBER/)
+    # if ( $rulename eq "SEPARATOR")
     # {
     #     printf "found!\n";
     # }
@@ -1814,6 +1816,8 @@ sub writeMarpaRule
     my $outputline          = "";
     my $outf                = $self->{outf};
 
+    my $printoutputcreated  = 1;
+
     SWITCH:
     {
         ##
@@ -1822,6 +1826,7 @@ sub writeMarpaRule
         ##
         ($self->testoption('buildkeywords') && $self->isKeywordLetter($rulename, $rule)) && do {
             $self->deleteMarpaRule($rulename);
+            $printoutputcreated = 0;
             last SWITCH;
         };
         #(exists $rule->{isFragmentOrChild} && $self->testoption('fragment2class') && fragmentEligible2Convert($rule)) && do {
@@ -1829,7 +1834,7 @@ sub writeMarpaRule
         #   conditionally translate eligible fragments to classes if the fragment is not tagged with a '?' quantifier
         ##
         ($self->testoption('fragment2class') && $rulename !~ /^opt_/ && fragmentEligible2Convert($rule)) && do {
-            $self->convertFragment2Class($rulename, $rule, $options );
+            $self->writeFragmentAsClass($rulename, $rule, $options );
             last SWITCH;
         };
         ##
@@ -1843,12 +1848,15 @@ sub writeMarpaRule
         #   print a simple rule
         ##
         do {
-            $outputline .= $self->printRhs( $rulename, $rule, $options );
+            $outputline = $self->printRhs( $rulename, $rule, $options );
+            $printoutputcreated = 0 if !length($outputline);
             last SWITCH;
         };
     }
 
-    printf $outf "%s\n\n", $outputline if length($outputline) > 0;
+    printf $outf "%s\n", $outputline if length($outputline) > 0;
+
+    return $printoutputcreated;
 }
 
 sub dumpMarpaRule
@@ -1965,6 +1973,9 @@ sub writeMarpaGrammar
     #   print rules in the order of the input file
     ##
     my $status = 0;
+    my $subrulestatus = 0;
+    my $rulegeneratedoutput = 1;
+
     for my $rule (@$ruletable)
     {
         if (exists $rule->{comment} && !$self->testoption('stripallcomments'))
@@ -1988,17 +1999,28 @@ sub writeMarpaGrammar
             $status = 1;
         }
 
-        $self->writeMarpaRule($rulename);
+        ##
+        #   create a block of text for a rule with its subrules
+        ##
+        printf $outf "\n" if exists $subrules->{$rulename} && !$subrulestatus && $rulegeneratedoutput;
+        $subrulestatus = 0;
+
+        $rulegeneratedoutput = $self->writeMarpaRule($rulename);
 
         ##
         #   print the generated subrules immediately after the parent rule
         ##
         if (exists $subrules->{$rulename})
         {
+            my $subrulegeneratedoutput = 0;
+
             for my $subrulename (@{$subrules->{$rulename}})
             {
-                $self->writeMarpaRule($subrulename);
+                $subrulegeneratedoutput += $self->writeMarpaRule($subrulename);
             }
+
+            printf $outf "\n" if $subrulegeneratedoutput;
+            $subrulestatus = 1;
         }
     }
 
@@ -2064,7 +2086,7 @@ $generator->shiftlazytogreedy   if exists $options->{g};
 $generator->buildkeywords       if exists $options->{k};
 $generator->stripactions        if exists $options->{p};
 $generator->setVerbosity(2)     if exists $options->{t};
-$generator->splitliteral        if exists $options->{u} || exists $options->{k};
+$generator->matchcaseinsensitive        if exists $options->{u} || exists $options->{k};
 
 my $outputfile  = '-';
 $outputfile     = $options->{o} if exists $options->{o};
